==> data-platform-telemetry-0.5.0-20220413.152444-5.jar: signal/data_platform_telemetry/metrics.clj <==
(ns signal.data-platform-telemetry.metrics
  (:require [iapetos.core :as prometheus]
            [iapetos.collector.jvm :as jvm]
            [iapetos.collector.ring :as ring]
            [java-time.convert :refer [to-millis-from-epoch]]))

(defn- ms->seconds [time]
  (/ time 1000.0))

(defn take-while-and-n-more
  [pred n coll]
  (lazy-seq
   (when-let [s (seq coll)]
     (if (pred (first s))
       (cons (first s) (take-while-and-n-more pred n (rest s)))
       (take n s)))))

(defn exponential-buckets
  "Generates exponential buckets based on <exponent>, starting at <start>,
   and ending with the bucket including the <max-value> - this may be larger than the max-value to include it."
  [start factor max-value]
  (loop [results [start]]
    (let [previous (last results)]
      (if (>= previous max-value)
        results
        (recur (conj results (* factor previous)))))))

(def ^:private default-processing-stage-buckets-seconds [0.01 0.025 0.05 0.075 0.1 0.25 0.5 1.0 2.0])
(def ^:private default-results-stage-buckets-seconds [0.1 0.25 0.5 0.75 1.0 2.0 4.0 8.0 16.0])

(defn get-default-histogram-buckets
  "Returns the default buckets for histograms for the metrics library.

   Buckets can subsequently be overwritten by overwriting a given key."
  [{:keys [max-batch-size] :or {max-batch-size 50}}]
  {:message-read-time default-processing-stage-buckets-seconds
   :message-process-time default-results-stage-buckets-seconds
   :message-total-time default-results-stage-buckets-seconds
   :message-batch-size (cond
                         (= 1 0) [0 1]
                         (<= max-batch-size 50) (take-while #(<= max-batch-size %) (map #(* 2 %) (range)))
                         :else (exponential-buckets 25 2 max-batch-size))

   :message-result-write-time default-processing-stage-buckets-seconds
   :message-publish-time default-processing-stage-buckets-seconds
   :message-ack-time default-processing-stage-buckets-seconds

   ;; needs to be set by user
   :pipeline-document-latency nil})

(comment
  (exponential-buckets 25 2 50000))


(def ^:dynamic *default-registry* prometheus/default-registry)

;; the histogram keywords contain fake namespaces to make sure they are not prefixed by "default-"
;; note that counters must end with _total to be valid in future Prometheus versions (i.e. under OpenMetrics)
(defn init
  "Initialises a Prometheus metrics registry and returns it."
  ([{:keys [registry
            ring?
            jvm?] :or {ring? true
                       jvm? true
                       registry *default-registry*}}]
   (-> registry
       (cond->
        jvm? jvm/initialize
        ring? ring/initialize)))
  ([]
   (init nil)))

(defn queue-reader
  "Registers queue reader metrics on the given registry."
  [registry {:keys [histogram-bucket-overrides]
             :as config}]
  (let [histogram-buckets (merge (get-default-histogram-buckets config) histogram-bucket-overrides)]
    (-> registry
        (prometheus/register-lazy
         (prometheus/histogram :read/time
                               {:buckets (:message-read-time histogram-buckets)
                                :description "The time taken to read the data for the message from S3."})
         (prometheus/histogram :ack/time
                               {:buckets (:message-ack-time histogram-buckets)
                                :description "The time taken to acknowledge a message in seconds."}))
        (prometheus/register
         (prometheus/histogram :message/process-time
                               {:buckets (:message-process-time histogram-buckets)
                                :description "The time taken to run the core process on the message in seconds."})
         (prometheus/histogram :message/total-time
                               {:buckets (:message-total-time histogram-buckets)
                                :description "The total time for message processing in seconds."})
         (prometheus/histogram :message/batch-size
                               {:buckets (:message-batch-size histogram-buckets)
                                :description "The number of documents in the message."})
         (prometheus/gauge :queue-worker/messages-in-flight {:description "The number of messages in flight being processed."})
         (prometheus/counter :queue-worker/errors-total {:description "Counter of errors occurred when processing messages."})))))

(defn queue-writer
  "Registers queue worker metrics on the given registry."
  [registry {:keys [histogram-bucket-overrides]
             :as config}]
  (let [histogram-buckets (merge (get-default-histogram-buckets config) histogram-bucket-overrides)]
    (-> registry
        (prometheus/register
         (prometheus/histogram :write/time
                               {:buckets (:message-result-write-time histogram-buckets)
                                :description "The time taken to write the output to S3."})
         (prometheus/histogram :sqs/write-time
                               {:buckets (:message-publish-time histogram-buckets)
                                :description "The time taken to publish the output message in seconds. Named sqs for legacy compatibility."})))))

(def ^:private high-watermark-time-seconds
  (atom {}))

(defn pipeline-latency
  "Registers metrics to track latency from point-to-point in a pipeline to the given registry.

   The configuration histogram-bucket-overrides.pipeline-document-latency must be provided - be sure to include the explicit value of any SLOs in your bucket distribution."
  [registry {:keys [histogram-bucket-overrides]
             :as config}]
  (assert (seq (:pipeline-document-latency histogram-bucket-overrides)) "Field :pipeline-document-latency should be provided in :histogram-bucket-overrides")
  (let [histogram-buckets (merge (get-default-histogram-buckets config) histogram-bucket-overrides)
        registry (-> registry
                     (prometheus/register
                      (prometheus/histogram :pipeline/document-latency-seconds {:description "A histogram of the age of the document from when Signal received the data. (time it was received by Signal - time it was output by the Sink)."
                                                                                :buckets (:pipeline-document-latency histogram-buckets)
                                                                                :labels [:measured_from
                                                                                         :mode]})

                      (prometheus/gauge :pipeline/document-latency-high-water-mark-seconds {:description "The received epoch time in seconds of the last outputted document."
                                                                                            :labels [:measured_from
                                                                                                     :mode]})))]

    (reset! high-watermark-time-seconds {})
    (add-watch high-watermark-time-seconds :record-pipeline-document-processed-high-water-mark-age-seconds
               (fn [_ _ _ val]
                 (doseq [[[mode measured_from] value] val]
                   (prometheus/set registry :pipeline/document-latency-high-water-mark-seconds {:mode mode
                                                                                                :measured_from measured_from} value))))
    registry))

(defn observe-time-seconds
  "Given a metric name and a start/end timestamp in milliseconds, observes the time in seconds between the two times."
  [metric registry start-timestamp-ms end-timestamp-ms]
  (try
    (prometheus/observe registry metric (ms->seconds (- end-timestamp-ms start-timestamp-ms)))
    (catch Exception e
      (throw (ex-info "Failed to observe metric" {:metric metric} e)))))


(defmacro message-read-time
  "Track the time taken by the macro body to read data for the message from storage, e.g S3."
  [registry & body]
  `(prometheus/with-duration (~registry :read/time)
     ~@body))

(def observe-message-read-time
  "Track the time taken to read data for the message from storage, e.g S3."
  (partial observe-time-seconds :read/time))


(defmacro message-result-write-time
  "Track the time taken by the macro body to write the message result to storage, e.g S3."
  [registry & body]
  `(prometheus/with-duration (~registry :write/time)
     ~@body))

(def observe-message-result-write-time
  "Track the time taken to write the message result to storage, e.g S3."
  (partial observe-time-seconds :write/time))

(defmacro message-publish-time
  "Track the time taken by the macro body to publish a new onward message, e.g to SNS."
  [registry & body]
  `(prometheus/with-duration (~registry :sqs/write-time)
     ~@body))

(def observe-message-publish-time
  "Track the time taken to publish a new onward message, e.g to SNS."
  (partial observe-time-seconds :sqs/write-time))

(defmacro message-ack-time
  "Track the total time taken by the macro body to acknowledge (i.e. output) the message result, e.g. writing to S3 and publishing to SNS."
  [registry & body]
  `(prometheus/with-duration (~registry :ack/time)
     ~@body))

(def observe-message-ack-time
  "Track the total time taken to acknowledge (i.e. output) the message result, e.g. writing to S3 and publishing to SNS."
  (partial observe-time-seconds :ack/time))


(defmacro message-process-time
  "Track the time taken by the macro body to run the service logic on the message."
  [registry & body]
  `(prometheus/with-duration (~registry :message/process-time)
     ~@body))

(def observe-message-process-time
  "Track the time taken to run the service logic on the message."
  (partial observe-time-seconds :message/process-time))


(defmacro message-total-time
  "Track the total time taken by the macro body to run the read the message, run logic, and acknoweledge it."
  [registry & body]
  `(prometheus/with-duration (~registry :message/total-time)
     ~@body))

(def observe-message-total-time
  "Track the total time taken to run the read the message, run logic, and acknoweledge it."
  (partial observe-time-seconds :message/total-time))


(defn increment-messages-in-flight
  "Increment the number of messages currently being processed by the service."
  ([registry n]
   (prometheus/inc registry :queue-worker/messages-in-flight n))
  ([registry]
   (increment-messages-in-flight registry 1)))

(defn decrement-messages-in-flight
  "Decrement the number of messages currently being processed by the service."
  ([registry n]
   (prometheus/dec registry :queue-worker/messages-in-flight n))
  ([registry]
   (decrement-messages-in-flight registry 1)))


(defn observe-message-batch-size
  "Observe the size of a batch being processed, e.g. 50 for an article mini-batch, or 1 for a single document."
  [registry size]
  (prometheus/observe registry :message/batch-size size))

(defn observe-queue-worker-error
  "Observe an error in processing."
  [registry _]
  (prometheus/inc registry :queue-worker/errors-total))


(defn- calc-age-seconds [time-seconds]
  (- (ms->seconds (System/currentTimeMillis)) time-seconds))

(defn record-pipeline-latency
  "Observe pipeline latency metrics for a document.

   These metrics are labelled by
   - \"mode\": the mode of operation of the service if applicable, e.g. realtime, non-realtime
   - \"measured-from\": where the latency was measured from, e.g. \"ingestion\" or \"pipeline_start\""
  [registry {:keys [mode
                    measured-from]} document-received-date-time]
  (let [start-time-epoch-seconds (quot (to-millis-from-epoch document-received-date-time)
                                       1000)]
    (prometheus/observe registry :pipeline/document-latency-seconds {:mode mode
                                                                     :measured_from measured-from} (calc-age-seconds start-time-epoch-seconds))
    ;; do not set a high watermark in the future
    (when (>= (/ (System/currentTimeMillis) 1000) start-time-epoch-seconds)
      (swap! high-watermark-time-seconds update [mode measured-from] #(max start-time-epoch-seconds (or % 0))))))


(defn ring-wrap-instrumentation
  "Wraps a ring application in Prometheus instrumentation."
  [registry app]
  (ring/wrap-instrumentation app
                             registry
                             {:path-fn (fn [_] "/")}))

(defn ring-wrap-metrics-expose
  "Creates a ring handler which exposes Prometheus metrics for the given registry on /internal/metrics."
  [registry app]
  (ring/wrap-metrics-expose app
                            registry
                            {:path "/internal/metrics"}))

==> data-platform-telemetry-0.5.0-20220505.145621-12.jar: signal/data_platform_telemetry/metrics.clj <==
(ns signal.data-platform-telemetry.metrics
  (:require [clojure.string :as string]
            [iapetos.core :as prometheus]
            [iapetos.collector.jvm :as jvm]
            [iapetos.collector.ring :as ring]
            [java-time.convert :refer [to-millis-from-epoch]]))

(defn- ms->seconds [time]
  (/ time 1000.0))

(defn take-while-and-n-more
  [pred n coll]
  (lazy-seq
   (when-let [s (seq coll)]
     (if (pred (first s))
       (cons (first s) (take-while-and-n-more pred n (rest s)))
       (take n s)))))

(defn exponential-buckets
  "Generates buckets with exponentially growing sizes based on <factor>, with initial step <initial-step>, starting at <start>,
   and ending with the bucket including the <max-value> - this may be larger than the max-value to include it.

   If initial step is not provided, <start> is used as initial step."
  ([start factor max-value initial-step]
   (loop [previous-results [start]
          previous-step initial-step]
     (let [previous (last previous-results)
           step (* factor previous-step)
           result (+ previous step)
           results (conj previous-results result)]
       (if (< result max-value)
         (recur results step)
         results))))
  ([start factor max-value]
   (exponential-buckets start factor max-value start)))

(def ^:private default-processing-stage-buckets-seconds [0.01 0.025 0.05 0.075 0.1 0.25 0.5 1.0 2.0])
(def ^:private default-results-stage-buckets-seconds [0.1 0.25 0.5 0.75 1.0 2.0 4.0 8.0 16.0])

(defn- batch-size-buckets [max-batch-size]
  (cond
    (= 1 max-batch-size) [0 1]
    (<= max-batch-size 50) (into [] (take-while-and-n-more #(< % max-batch-size) 1 (map #(* 2 %) (range))))
    :else (mapv #(Math/ceil %) (exponential-buckets 10 1.1 max-batch-size))))

(defn get-default-histogram-buckets
  "Returns the default buckets for histograms for the metrics library.

   Buckets can subsequently be overwritten by overwriting a given key."
  [{:keys [max-batch-size] :or {max-batch-size 50}}]
  {:message-read-time default-processing-stage-buckets-seconds
   :message-process-time default-results-stage-buckets-seconds
   :message-total-time default-results-stage-buckets-seconds
   :message-batch-size (batch-size-buckets max-batch-size)

   :message-result-write-time default-processing-stage-buckets-seconds
   :message-publish-time default-processing-stage-buckets-seconds
   :message-ack-time default-processing-stage-buckets-seconds

   ;; needs to be set by user
   :pipeline-document-latency nil})


(defn pipeline-document-latency-histogram-buckets
  "Generate buckets for the pipeline document latency histogram.

   This function should be used in the REPL to generate latency histogram buckets
   "
  ([{:keys [target-slo
            low-resolution-step
            high-resolution-step
            max-bucket-value
            exponential-factor]
     :or {target-slo 90
          low-resolution-step 15
          high-resolution-step 5
          max-bucket-value (* 180 60)
          exponential-factor 1.2}}]
   (let [high-resolution-floor (/ target-slo 2)
         high-resolution-ceiling (+ target-slo (/ target-slo 2))
         low-resolution-for-low-values (range 0 high-resolution-floor low-resolution-step)
         high-resolution-around-slo (range high-resolution-floor high-resolution-ceiling high-resolution-step)
         exponential-for-high-values (rest (exponential-buckets (last high-resolution-around-slo) exponential-factor max-bucket-value))
         document-latency-buckets (concat low-resolution-for-low-values
                                          high-resolution-around-slo
                                          exponential-for-high-values)]
     (->> document-latency-buckets
          (map #(format "%.1f" (double %)))
          (string/join ","))))
  ([]
   (pipeline-document-latency-histogram-buckets {})))

(comment
  (pipeline-document-latency-histogram-buckets))


(comment
  (exponential-buckets 25 2 50000)
  (batch-size-buckets 1)
  (batch-size-buckets 50)
  (batch-size-buckets 1000))


(def ^:dynamic *default-registry* prometheus/default-registry)

;; the histogram keywords contain fake namespaces to make sure they are not prefixed by "default-"
;; note that counters must end with _total to be valid in future Prometheus versions (i.e. under OpenMetrics)
(defn init
  "Initialises a Prometheus metrics registry and returns it."
  ([{:keys [registry
            ring?
            jvm?] :or {ring? true
                       jvm? true
                       registry *default-registry*}}]
   (-> registry
       (cond->
        jvm? jvm/initialize
        ring? ring/initialize)))
  ([]
   (init nil)))

(defn queue-reader
  "Registers queue reader metrics on the given registry."
  [registry {:keys [histogram-bucket-overrides]
             :as config}]
  (let [histogram-buckets (merge (get-default-histogram-buckets config) histogram-bucket-overrides)]
    (-> registry
        (prometheus/register-lazy
         (prometheus/histogram :read/time
                               {:buckets (:message-read-time histogram-buckets)
                                :description "The time taken to read the data for the message from S3."})
         (prometheus/histogram :ack/time
                               {:buckets (:message-ack-time histogram-buckets)
                                :description "The time taken to acknowledge a message in seconds."}))
        (prometheus/register
         (prometheus/histogram :message/process-time
                               {:buckets (:message-process-time histogram-buckets)
                                :description "The time taken to run the core process on the message in seconds."})
         (prometheus/histogram :message/total-time
                               {:buckets (:message-total-time histogram-buckets)
                                :description "The total time for message processing in seconds."})
         (prometheus/histogram :message/batch-size
                               {:buckets (:message-batch-size histogram-buckets)
                                :description "The number of documents in the message."})
         (prometheus/gauge :queue-worker/messages-in-flight {:description "The number of messages in flight being processed."})
         (prometheus/counter :queue-worker/errors-total {:description "Counter of errors occurred when processing messages."})))))

(defn queue-writer
  "Registers queue worker metrics on the given registry."
  [registry {:keys [histogram-bucket-overrides]
             :as config}]
  (let [histogram-buckets (merge (get-default-histogram-buckets config) histogram-bucket-overrides)]
    (-> registry
        (prometheus/register
         (prometheus/histogram :write/time
                               {:buckets (:message-result-write-time histogram-buckets)
                                :description "The time taken to write the output to S3."})
         (prometheus/histogram :sqs/write-time
                               {:buckets (:message-publish-time histogram-buckets)
                                :description "The time taken to publish the output message in seconds. Named sqs for legacy compatibility."})))))

(def ^:private high-watermark-time-seconds
  (atom {}))

(defn pipeline-latency
  "Registers metrics to track latency from point-to-point in a pipeline to the given registry.

   The configuration histogram-bucket-overrides.pipeline-document-latency must be provided - be sure to include the explicit value of any SLOs in your bucket distribution."
  [registry {:keys [histogram-bucket-overrides]
             :as config}]
  (assert (seq (:pipeline-document-latency histogram-bucket-overrides)) "Field :pipeline-document-latency should be provided in :histogram-bucket-overrides")
  (let [histogram-buckets (merge (get-default-histogram-buckets config) histogram-bucket-overrides)
        registry (-> registry
                     (prometheus/register
                      (prometheus/histogram :pipeline/document-latency-seconds {:description "A histogram of the age of the document from when Signal received the data. (time it was received by Signal - time it was output by the Sink)."
                                                                                :buckets (:pipeline-document-latency histogram-buckets)
                                                                                :labels [:measured_from
                                                                                         :mode]})

                      (prometheus/gauge :pipeline/document-latency-high-water-mark-seconds {:description "The received epoch time in seconds of the last outputted document."
                                                                                            :labels [:measured_from
                                                                                                     :mode]})))]

    (reset! high-watermark-time-seconds {})
    (add-watch high-watermark-time-seconds :record-pipeline-document-processed-high-water-mark-age-seconds
               (fn [_ _ _ val]
                 (doseq [[[mode measured_from] value] val]
                   (prometheus/set registry :pipeline/document-latency-high-water-mark-seconds {:mode mode
                                                                                                :measured_from measured_from} value))))
    registry))

(defn observe-time-seconds
  "Given a metric name and a start/end timestamp in milliseconds, observes the time in seconds between the two times."
  [metric registry start-timestamp-ms end-timestamp-ms]
  (try
    (prometheus/observe registry metric (ms->seconds (- end-timestamp-ms start-timestamp-ms)))
    (catch Exception e
      (throw (ex-info "Failed to observe metric" {:metric metric} e)))))


(defmacro message-read-time
  "Track the time taken by the macro body to read data for the message from storage, e.g S3."
  [registry & body]
  `(prometheus/with-duration (~registry :read/time)
     ~@body))

(def observe-message-read-time
  "Track the time taken to read data for the message from storage, e.g S3."
  (partial observe-time-seconds :read/time))


(defmacro message-result-write-time
  "Track the time taken by the macro body to write the message result to storage, e.g S3."
  [registry & body]
  `(prometheus/with-duration (~registry :write/time)
     ~@body))

(def observe-message-result-write-time
  "Track the time taken to write the message result to storage, e.g S3."
  (partial observe-time-seconds :write/time))

(defmacro message-publish-time
  "Track the time taken by the macro body to publish a new onward message, e.g to SNS."
  [registry & body]
  `(prometheus/with-duration (~registry :sqs/write-time)
     ~@body))

(def observe-message-publish-time
  "Track the time taken to publish a new onward message, e.g to SNS."
  (partial observe-time-seconds :sqs/write-time))

(defmacro message-ack-time
  "Track the total time taken by the macro body to acknowledge (i.e. output) the message result, e.g. writing to S3 and publishing to SNS."
  [registry & body]
  `(prometheus/with-duration (~registry :ack/time)
     ~@body))

(def observe-message-ack-time
  "Track the total time taken to acknowledge (i.e. output) the message result, e.g. writing to S3 and publishing to SNS."
  (partial observe-time-seconds :ack/time))


(defmacro message-process-time
  "Track the time taken by the macro body to run the service logic on the message."
  [registry & body]
  `(prometheus/with-duration (~registry :message/process-time)
     ~@body))

(def observe-message-process-time
  "Track the time taken to run the service logic on the message."
  (partial observe-time-seconds :message/process-time))


(defmacro message-total-time
  "Track the total time taken by the macro body to run the read the message, run logic, and acknoweledge it."
  [registry & body]
  `(prometheus/with-duration (~registry :message/total-time)
     ~@body))

(def observe-message-total-time
  "Track the total time taken to run the read the message, run logic, and acknoweledge it."
  (partial observe-time-seconds :message/total-time))


(defn increment-messages-in-flight
  "Increment the number of messages currently being processed by the service."
  ([registry n]
   (prometheus/inc registry :queue-worker/messages-in-flight n))
  ([registry]
   (increment-messages-in-flight registry 1)))

(defn decrement-messages-in-flight
  "Decrement the number of messages currently being processed by the service."
  ([registry n]
   (prometheus/dec registry :queue-worker/messages-in-flight n))
  ([registry]
   (decrement-messages-in-flight registry 1)))


(defn observe-message-batch-size
  "Observe the size of a batch being processed, e.g. 50 for an article mini-batch, or 1 for a single document."
  [registry size]
  (prometheus/observe registry :message/batch-size size))

(defn observe-queue-worker-error
  "Observe an error in processing."
  [registry _]
  (prometheus/inc registry :queue-worker/errors-total))


(defn- calc-age-seconds [time-seconds]
  (- (ms->seconds (System/currentTimeMillis)) time-seconds))

(defn record-pipeline-latency
  "Observe pipeline latency metrics for a document.

   These metrics are labelled by
   - \"mode\": the mode of operation of the service if applicable, e.g. realtime, non-realtime
   - \"measured-from\": where the latency was measured from, e.g. \"ingestion\" or \"pipeline_start\""
  [registry {:keys [mode
                    measured-from]} document-received-date-time]
  (let [start-time-epoch-seconds (quot (to-millis-from-epoch document-received-date-time)
                                       1000)]
    (prometheus/observe registry :pipeline/document-latency-seconds {:mode mode
                                                                     :measured_from measured-from} (calc-age-seconds start-time-epoch-seconds))
    ;; do not set a high watermark in the future
    (when (>= (/ (System/currentTimeMillis) 1000) start-time-epoch-seconds)
      (swap! high-watermark-time-seconds update [mode measured-from] #(max start-time-epoch-seconds (or % 0))))))


(defn ring-wrap-instrumentation
  "Wraps a ring application in Prometheus instrumentation."
  [app registry]
  (ring/wrap-instrumentation app
                             registry
                             {:path-fn (fn [_] "/")}))

(defn ring-wrap-metrics-expose
  "Creates a ring handler which exposes Prometheus metrics for the given registry on /internal/metrics."
  [app registry]
  (ring/wrap-metrics-expose app
                            registry
                            {:path "/internal/metrics"}))
